@@ -927,10 +927,17 @@ static int has_smi_counter(void)
 	case 0x56:      /* BDX-DE */
 	case 0x4E:      /* SKL */
 	case 0x5E:      /* SKL */
+	case 0x8E:      /* KBL */
+	case 0x9E:      /* KBL */
+	case 0x55:      /* SKX */
 	case 0x37:      /* BYT */
 	case 0x4D:      /* AVN */
 	case 0x4C:      /* AMT */
 	case 0x57:      /* PHI */
+	case 0x5C:      /* BXT */
+	case 0x5F:	/* DNV */
+	case 0x7A:	/* Gemini Lake */
+	case 0x85:	/* Knights Mill */
 		break;
 	default:
 		return 0;
@@ -1237,6 +1244,17 @@ static void *timerthread(void *param)
 	}
 
 out:
+	if (refresh_on_max) {
+		pthread_mutex_lock(&refresh_on_max_lock);
+		/* We could reach here with both shutdown and allstopped unset (0).
+		 * Set shutdown with synchronization to notify the main
+		 * thread not to be blocked when it should exit.
+		 */
+		shutdown++;
+		pthread_cond_signal(&refresh_on_max_cond);
+		pthread_mutex_unlock(&refresh_on_max_lock);
+	}
+
 	if (par->mode == MODE_CYCLIC)
 		timer_delete(timer);
 
@@ -1317,7 +1335,6 @@ static void display_help(int error)
 	       "-m       --mlockall        lock current and future memory allocations\n"
 	       "-M       --refresh_on_max  delay updating the screen until a new max\n"
 	       "			   latency is hit. Userful for low bandwidth.\n"
-	       "-n       --nanosleep       use clock_nanosleep\n"
 	       "	 --notrace	   suppress tracing\n"
 	       "-N       --nsecs           print results in ns instead of us (default us)\n"
 	       "-o RED   --oscope=RED      oscilloscope mode, reduce verbose output by RED\n"
@@ -1360,7 +1377,8 @@ static void display_help(int error)
 	       "                           format: n:c:v n=tasknum c=count v=value in us\n"
 	       "-w       --wakeup          task wakeup tracing (used with -b)\n"
 	       "-W       --wakeuprt        rt task wakeup tracing (used with -b)\n"
-	       "	 --dbg_cyclictest  print info useful for debugging cyclictest\n",
+	       "	 --dbg_cyclictest  print info useful for debugging cyclictest\n"
+	       "-x	 --posix_timers    use POSIX timers instead of clock_nanosleep.\n",
 	       tracers
 		);
 	if (error)
@@ -1368,7 +1386,7 @@ static void display_help(int error)
 	exit(EXIT_SUCCESS);
 }
 
-static int use_nanosleep;
+static int use_nanosleep = MODE_CLOCK_NANOSLEEP;
 static int timermode = TIMER_ABSTIME;
 static int use_system;
 static int priority;
@@ -1489,6 +1507,7 @@ enum option_values {
 	OPT_TRIGGER_NODES, OPT_UNBUFFERED, OPT_NUMA, OPT_VERBOSE, OPT_WAKEUP,
 	OPT_WAKEUPRT, OPT_DBGCYCLIC, OPT_POLICY, OPT_HELP, OPT_NUMOPTS,
 	OPT_ALIGNED, OPT_SECALIGNED, OPT_LAPTOP, OPT_SMI, OPT_TRACEMARK,
+	OPT_POSIX_TIMERS,
 };
 
 /* Process commandline options */
@@ -1526,7 +1545,6 @@ static void process_options (int argc, char *argv[], int max_cpus)
 			{"loops",            required_argument, NULL, OPT_LOOPS },
 			{"mlockall",         no_argument,       NULL, OPT_MLOCKALL },
 			{"refresh_on_max",   no_argument,       NULL, OPT_REFRESH },
-			{"nanosleep",        no_argument,       NULL, OPT_NANOSLEEP },
 			{"nsecs",            no_argument,       NULL, OPT_NSECS },
 			{"oscope",           required_argument, NULL, OPT_OSCOPE },
 			{"traceopt",         required_argument, NULL, OPT_TRACEOPT },
@@ -1553,9 +1571,10 @@ static void process_options (int argc, char *argv[], int max_cpus)
 			{"dbg_cyclictest",   no_argument,       NULL, OPT_DBGCYCLIC },
 			{"policy",           required_argument, NULL, OPT_POLICY },
 			{"help",             no_argument,       NULL, OPT_HELP },
+			{"posix_timers",     no_argument,	NULL, OPT_POSIX_TIMERS },
 			{NULL, 0, NULL, 0}
 		};
-		int c = getopt_long(argc, argv, "a::A::b:Bc:Cd:D:Efh:H:i:Il:MnNo:O:p:PmqrRsSt::uUvD:wWT:",
+		int c = getopt_long(argc, argv, "a::A::b:Bc:Cd:D:Efh:H:i:Il:MNo:O:p:PmqrRsSt::uUvD:wWT:x",
 				    long_options, &option_index);
 		if (c == -1)
 			break;
@@ -1647,9 +1666,6 @@ static void process_options (int argc, char *argv[], int max_cpus)
 		case 'M':
 		case OPT_REFRESH:
 			refresh_on_max = 1; break;
-		case 'n':
-		case OPT_NANOSLEEP:
-			use_nanosleep = MODE_CLOCK_NANOSLEEP; break;
 		case 'N':
 		case OPT_NSECS:
 			use_nsecs = 1; break;
@@ -1684,7 +1700,6 @@ static void process_options (int argc, char *argv[], int max_cpus)
 		case 'R':
 		case OPT_RESOLUTION:
 			check_clock_resolution = 1; break;
-		case 's':
 		case OPT_SECALIGNED:
 			secaligned = 1;
 			if (optarg != NULL)
@@ -1694,6 +1709,7 @@ static void process_options (int argc, char *argv[], int max_cpus)
 			else
 				offset = 0;
 			break;
+		case 's':
 		case OPT_SYSTEM:
 			use_system = MODE_SYS_OFFSET; break;
 		case 'S':
@@ -1703,7 +1719,6 @@ static void process_options (int argc, char *argv[], int max_cpus)
 			smp = 1;
 			num_threads = max_cpus;
 			setaffinity = AFFINITY_USEALL;
-			use_nanosleep = MODE_CLOCK_NANOSLEEP;
 			break;
 		case 't':
 		case OPT_THREADS:
@@ -1742,7 +1757,6 @@ static void process_options (int argc, char *argv[], int max_cpus)
 #ifdef NUMA
 			num_threads = max_cpus;
 			setaffinity = AFFINITY_USEALL;
-			use_nanosleep = MODE_CLOCK_NANOSLEEP;
 #else
 			warn("cyclictest was not built with the numa option\n");
 			warn("ignoring --numa or -U\n");
@@ -1756,6 +1770,9 @@ static void process_options (int argc, char *argv[], int max_cpus)
 		case 'W':
 		case OPT_WAKEUPRT:
 			tracetype = WAKEUPRT; break;
+		case 'x':
+		case OPT_POSIX_TIMERS:
+			use_nanosleep = MODE_CYCLIC; break;
 		case '?':
 		case OPT_HELP:
 			display_help(0); break;
@@ -2032,6 +2049,13 @@ static void print_hist(struct thread_param *par[], int nthreads)
 			fprintf(fd, " # %05lu others", par[i]->stats->hist_overflow - par[i]->stats->num_outliers);
 		fprintf(fd, "\n");
 	}
+	if (smi) {
+		fprintf(fd, "# SMIs:");
+		for (i = 0; i < nthreads; i++)
+			fprintf(fd, " %05lu", par[i]->stats->smi_count);
+		fprintf(fd, "\n");
+	}
+
 	fprintf(fd, "\n");
 
 	if (use_histfile)
@@ -2372,6 +2396,9 @@ int main(int argc, char **argv)
 			/*  allocate memory for a stack on appropriate node */
 			stack = rt_numa_numa_alloc_onnode(stksize, node, i);
 
+			/* touch the stack pages to pre-fault them in */
+			memset(stack, 0, stksize);
+
 			/* set the thread's stack */
 			if (pthread_attr_setstack(&attr, stack, stksize))
 				fatal("failed to set stack addr for thread %d to 0x%x\n",
@@ -2508,8 +2535,9 @@ int main(int argc, char **argv)
 
 		if (refresh_on_max) {
 			pthread_mutex_lock(&refresh_on_max_lock);
-			pthread_cond_wait(&refresh_on_max_cond,
-					  &refresh_on_max_lock);
+			if (!shutdown)
+				pthread_cond_wait(&refresh_on_max_cond,
+						&refresh_on_max_lock);
 			pthread_mutex_unlock(&refresh_on_max_lock);
 		}
 	}
@@ -2519,6 +2547,9 @@ int main(int argc, char **argv)
 	shutdown = 1;
 	usleep(50000);
 
+	if (!verbose && !quiet && refresh_on_max)
+		printf("\033[%dB", num_threads + 2);
+
 	if (quiet)
 		quiet = 2;
 	for (i = 0; i < num_threads; i++) {
