@@ -20,9 +20,6 @@
  *
  * ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
  */
-#ifndef _GNU_SOURCE
-#define _GNU_SOURCE
-#endif
 #include <stdio.h>
 #ifndef __USE_XOPEN2K
 # define __USE_XOPEN2K
@@ -47,10 +44,6 @@
 
 #define gettid() syscall(__NR_gettid)
 
-#ifndef VERSION_STRING
-#define VERSION_STRING 0.3
-#endif
-
 int nr_tasks;
 int lfd;
 
@@ -116,6 +109,7 @@ static int nr_runs = NR_RUNS;
 static int prio_start = PRIO_START;
 static int check;
 static int stop;
+static int equal;
 
 static unsigned long long now;
 
@@ -178,15 +172,17 @@ static void usage(char **argv)
 		p--;
 	p++;
 
-	printf("%s %1.2f\n", p, VERSION_STRING);
+	printf("%s %1.2f\n", p, VERSION);
 	printf("Usage:\n"
 	       "%s <options> nr_tasks\n\n"
-	       "-p prio --prio  prio        base priority to start RT tasks with (2) \n"
+	       "-p prio --prio  prio        base priority to start RT tasks with (2)\n"
 	       "-r time --run-time time     Run time (ms) to busy loop the threads (20)\n"
 	       "-s time --sleep-time time   Sleep time (ms) between intervals (100)\n"
 	       "-m time --maxerr time       Max allowed error (microsecs)\n"
 	       "-l loops --loops loops      Number of iterations to run (50)\n"
+	       "-e                          Use equal prio for #CPU-1 tasks (requires > 2 CPUS)\n"
 	       "-c    --check               Stop if lower prio task is quicker than higher (off)\n"
+	       "-h    --help\n"
 	       "  () above are defaults \n",
 		p);
 	exit(0);
@@ -207,18 +203,19 @@ static void parse_options (int argc, char *argv[])
 			{"help", no_argument, NULL, '?'},
 			{NULL, 0, NULL, 0}
 		};
-		int c = getopt_long (argc, argv, "p:r:s:m:l:ch",
+		int c = getopt_long (argc, argv, "p:r:s:m:l:ech",
 			long_options, &option_index);
 		if (c == -1)
 			break;
 		switch (c) {
 		case 'p': prio_start = atoi(optarg); break;
 		case 'r':
-			run_interval = ms2nano(atoi(optarg));
+			run_interval = atoi(optarg);
 			break;
-		case 's': interval = ms2nano(atoi(optarg)); break;
+		case 's': interval = atoi(optarg); break;
 		case 'l': nr_runs = atoi(optarg); break;
 		case 'm': max_err = usec2nano(atoi(optarg)); break;
+		case 'e': equal = 1; break;
 		case 'c': check = 1; break;
 		case '?':
 		case 'h':
@@ -226,14 +223,7 @@ static void parse_options (int argc, char *argv[])
 			break;
 		}
 	}
-	if (nr_runs <= 0) {
-		fprintf(stderr, "Error: --loops argument is non-positive.  Exiting.\n");
-		exit(-1);
-	}
-	if (prio_start < 1 || prio_start > 99) {
-		fprintf(stderr, "Error: invalid value for --prio: %d (valid: 1-99)\n", prio_start);
-		exit(-1);
-	}
+
 }
 
 static unsigned long long get_time(void)
@@ -263,6 +253,12 @@ static void record_time(int id, unsigned long long time, unsigned long l)
 	intervals_loops[loop][id] = l;
 }
 
+static int calc_prio(int id)
+{
+	int prio = equal && id && (id < nr_tasks - 1) ? 1 : id;
+	return prio + prio_start;
+}
+
 static void print_results(void)
 {
 	int i;
@@ -313,7 +309,7 @@ static void print_results(void)
 	printf("Parent pid: %d\n", getpid());
 
 	for (t=0; t < nr_tasks; t++) {
-		printf(" Task %d (prio %d) (pid %ld):\n", t, t + prio_start,
+		printf(" Task %d (prio %d) (pid %ld):\n", t, calc_prio(t),
 			thread_pids[t]);
 		printf("   Max: %lld us\n", nano2usec(tasks_max[t]));
 		printf("   Min: %lld us\n", nano2usec(tasks_min[t]));
@@ -338,7 +334,7 @@ static unsigned long busy_loop(unsigned long long start_time)
 	do {
 		l++;
 		time = get_time();
-	} while ((time - start_time) < run_interval);
+	} while ((time - start_time) < RUN_INTERVAL);
 
 	return l;
 }
@@ -347,8 +343,9 @@ void *start_task(void *data)
 {
 	long id = (long)data;
 	unsigned long long start_time;
+	int prio = calc_prio(id);
 	struct sched_param param = {
-		.sched_priority = id + prio_start,
+		.sched_priority = prio,
 	};
 	int ret;
 	int high = 0;
@@ -363,6 +360,7 @@ void *start_task(void *data)
 		perr("getting affinity");
 
 	pid = gettid();
+	thread_pids[id] = pid;
 
 	/* Check if we are the highest prio task */
 	if (id == nr_tasks-1)
@@ -413,6 +411,7 @@ static int check_times(int l)
 			    intervals[l][i] > last_length ||
 			    (intervals_length[l][i] > last_length &&
 			     intervals_length[l][i] - last_length > max_err)) {
+				ftrace_write("Task %d FAILED\n", thread_pids[i]);
 				check = -1;
 				return 1;
 			}
@@ -543,8 +542,8 @@ int main (int argc, char **argv)
 
 
 
-	intv.tv_sec = nano2sec(interval);
-	intv.tv_nsec = interval % sec2nano(1);
+	intv.tv_sec = nano2sec(INTERVAL);
+	intv.tv_nsec = INTERVAL % sec2nano(1);
 
 	print_progress_bar(0);
 
@@ -553,6 +552,9 @@ int main (int argc, char **argv)
 	for (loop=0; loop < nr_runs; loop++) {
 		unsigned long long end;
 
+		/* Release the CPU so all can get to the next barrier */
+		nanosleep(&intv, NULL);
+
 		now = get_time();
 
 		ftrace_write("Loop %d now=%lld\n", loop, now);
